/**
 * Copyright (C) 2008-2012 AgroSense Foundation.
 *
 * AgroSense is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version.
 *
 * There are special exceptions to the terms and conditions of the GPLv3 as it
 * is applied to this software, see the FLOSS License Exception
 * <http://www.agrosense.eu/foss-exception.html>.
 *
 * AgroSense is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
 * A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * AgroSense. If not, see <http://www.gnu.org/licenses/>.
 */
package nl.cloudfarming.client.field.model;

import com.vividsolutions.jts.geom.Envelope;
import com.vividsolutions.jts.geom.Geometry;
import com.vividsolutions.jts.geom.Point;
import com.vividsolutions.jts.io.ParseException;
import com.vividsolutions.jts.operation.valid.IsValidOp;
import com.vividsolutions.jts.operation.valid.TopologyValidationError;
import java.awt.Image;
import java.awt.datatransfer.DataFlavor;
import java.awt.datatransfer.Transferable;
import java.awt.datatransfer.UnsupportedFlavorException;
import java.beans.PropertyChangeSupport;
import java.io.IOException;
import java.io.Serializable;
import java.util.Collections;
import java.util.HashMap;
import java.util.Map;
import java.util.Objects;
import java.util.logging.Level;
import java.util.logging.Logger;
import nl.cloudfarming.client.geoviewer.Geometrical;
import org.openide.util.ImageUtilities;
import org.openide.util.NbBundle.Messages;

/**
 * non persistent object for storing geometries and additional information
 *
 *
 * @author Timon Veenstra
 */
public class Shape implements Geometrical<Geometry>, Transferable, Cloneable, Serializable {

    private static final Logger LOG = Logger.getLogger("nl.cloudfarming.client.field.model.Shape");
    public static final DataFlavor DATA_FLAVOR = new DataFlavor(Shape.class, "Shape");
    private static long lastId = 0l;
    private Geometry geometry;
    private Map<String, String> stringAttributeMap = new HashMap<>();
    private Map<String, Double> integerAttributeMap = new HashMap<>();
    private String name;
    private Double value;
    private final long id;
    private String geometryType;
    private TopologyValidationError toplogyValidationError;
    private boolean autocorrected = false;
    private ImportType importType = ImportType.SKIP;
    public static final String PROP_IMPORT_TYPE = "importType";
    public static final String PROP_NAME = "name";
    public static final String PROP_VALUE = "value";
    private transient java.beans.PropertyChangeSupport propertyChangeSupport = new java.beans.PropertyChangeSupport(this);
    private static final Image ICON = ImageUtilities.loadImage("nl/cloudfarming/client/field/model/shape_icon_polygon.png");

    /**
     * Add PropertyChangeListener.
     *
     * @param listener
     */
    @Override
    public void addPropertyChangeListener(java.beans.PropertyChangeListener listener) {
        propertyChangeSupport.addPropertyChangeListener(listener);
    }

    /**
     * Remove PropertyChangeListener.
     *
     * @param listener
     */
    @Override
    public void removePropertyChangeListener(java.beans.PropertyChangeListener listener) {
        propertyChangeSupport.removePropertyChangeListener(listener);
    }

    /**
     * @deprecated FIXME: only used by test classes, replace with mocks or use
     * the other constructor with a mocked Geometry.
     */
    @Deprecated
    public Shape() {
        this.id = lastId++;
    }

    /**
     * Construct a shape object based on a WKT string
     *
     * @param Geometry
     * @throws ParseException
     */
    public Shape(Geometry geometry) {
        this.id = lastId++;
        this.geometry = Objects.requireNonNull(geometry);

        // Determine and set shape type
        this.geometryType = geometry.getGeometryType().toUpperCase();
        // TODO: see if the validation methods in GeometryTools are still needed (if not, we can use a single branch for all types)
        // the polygonizer it uses will detect 3 types of errors, and returns all errors it found;
        // the IsValidOp handles all geometry types (therefore more error types), but will only return the first error found.
        // update: choice has been made easier, the polygonizer seems to miss at least one validation type (causing a
        // TopologyException later on in our processes), which does get caught by the IsValidOp, so always use the latter.

        IsValidOp validator = new IsValidOp(geometry);
        toplogyValidationError = validator.getValidationError();
        if (toplogyValidationError != null) {
            LOG.log(Level.FINEST, "topology validation error: {0}", toplogyValidationError.getMessage());
            
            Geometry correctedGeometry = AutocorrectGeometries.autocorrect(geometry);
            if (correctedGeometry != null) {
                // maybe later: keep the original geometry so we can visually compare them if individual shapes will get (multiview) topcomponents;
                // error details could be shown there as well, or in an aggregate report tab for the full shapefile.
                this.geometry = correctedGeometry;
                autocorrected = true;
            }
        }

        switch (geometryType) {
            case "MULTIPOLYGON":
                importType = ImportType.FIELD;
                break;
            case "POLYGON":
                importType = ImportType.FIELD;
                break;


            // TODO: Reimplement Helplines in appication, and restore the 2 actions beneath.
            /*
             * case "LINESTRING": importType = ImportType.HELPLINE; break; case
             * "MULTILINESTRING": importType = ImportType.HELPLINE; break;
             *
             */
            default:
                importType = null;
                break;
        }
    }

    public Double getValue() {
        return value;
    }

    public void setValue(Double value) {
        Double old = this.value;
        this.value = value;
        propertyChangeSupport.firePropertyChange(PROP_VALUE, old, value);
    }

    @Deprecated
    public ImportType getImportType() {
        return importType;
    }

    @Deprecated
    public void setImportType(ImportType importType) {
        ImportType old = this.importType;
        this.importType = importType;
        propertyChangeSupport.firePropertyChange(PROP_IMPORT_TYPE, old, importType);
    }

    public String getName() {
        return name;
    }

    public void setName(String name) {
        String old = this.name;
        this.name = name;
        propertyChangeSupport.firePropertyChange(PROP_NAME, old, name);
    }

    /**
     * @see nl.cloudfarming.client.geoviewer.LayerObject#getGeometry()
     * @return The Geometry of type MultiPolygon
     */
    @Override
    public Geometry getGeometry() {
        return geometry;
    }

    /**
     * @see nl.cloudfarming.client.geoviewer.LayerObject#getTooltipText()
     * @return TooltipText as String
     */
    @Override
    public String getTooltipText() {
        return new StringBuilder().append(getName()).append(" - ").append(getValue()).toString();
    }

    /**
     * @see nl.cloudfarming.client.geoviewer.LayerObject#getId()
     * @return The Id of the wrapped object
     */
    public long getId() {
        return id;
    }

    @Override
    public Object clone() throws CloneNotSupportedException {
        return super.clone();
    }

    @Override
    public boolean equals(Object obj) {
        if (obj == null) {
            return false;
        }
        if (getClass() != obj.getClass()) {
            return false;
        }
        final Shape other = (Shape) obj;
        if (this.id != other.id) {
            return false;
        }
        return true;
    }

    @Override
    public int hashCode() {
        int hash = super.hashCode();
        hash = 79 * hash + (int) (this.id ^ (this.id >>> 32));
        return hash;
    }

    @Override
    public Geometry getRenderObject(Envelope envelope) {
        if (envelope.intersects(geometry.getEnvelopeInternal())) {
            return geometry;
        }
        return null;
    }

    @Override
    public Point getCentroid() {
        return this.geometry.getCentroid();
    }

    @Override
    public DataFlavor[] getTransferDataFlavors() {
        return new DataFlavor[]{DATA_FLAVOR};
    }

    @Override
    public boolean isDataFlavorSupported(DataFlavor flavor) {
        return DATA_FLAVOR.equals(flavor);
    }

    @Override
    public Object getTransferData(DataFlavor flavor) throws UnsupportedFlavorException, IOException {
        if (isDataFlavorSupported(flavor)) {
            return this;
        }
        throw new UnsupportedFlavorException(flavor);
    }

    /**
     * Add a string attribute to this shape a string attribute is additional
     * data that is contained in the geometry file
     *
     * @param name the name of the attribute
     * @param value the value of the attribute
     */
    public void addStringAttribute(String name, String value) {
        stringAttributeMap.put(name, value);
    }

    /**
     * Gets a unmodifiable map of the Attributes of this shape This map contains
     * additional information of the shape for instance it could contain the
     * name of the shape
     *
     * @return the map
     */
    public Map<String, String> getStringAttributeMap() {
        return Collections.unmodifiableMap(stringAttributeMap);
    }

    /**
     * Add a string attribute to this shape a string attribute is additional
     * data that is contained in the geometry file
     *
     * @param name the name of the attribute
     * @param value the value of the attribute
     */
    public void addIntegerAttribute(String name, Double value) {
        integerAttributeMap.put(name, value);
    }

    /**
     * Gets a unmodifiable map of the Attributes of this shape This map contains
     * additional information of the shape for instance it could contain the
     * name of the shape
     *
     * @return the map
     */
    public Map<String, Double> getIntegerAttributeMap() {
        return Collections.unmodifiableMap(integerAttributeMap);
    }

    private void readObject(java.io.ObjectInputStream in)
            throws IOException, ClassNotFoundException {
        in.defaultReadObject();
        propertyChangeSupport = new PropertyChangeSupport(this);
    }

    /**
     * Set the name of the shape based on a attribute that should be present in
     * the stringAttibute map if the map contains attribute name will set the
     * name of this shape to the value in the map if the parsed attributeName is
     * is not found in the map will do nothing and log a warning
     *
     * @param attributeName the key in the attributeMap
     */
    public void setNameFromAttribute(String attributeName) {
        String theValue = stringAttributeMap.get(attributeName);

        String intValue = integerAttributeMap.get(attributeName) != null ? integerAttributeMap.get(attributeName).toString() : "";
        if (theValue != null && !theValue.equals("")) {
            setName(theValue);
        } else if (intValue != null && !intValue.equals("")) {
            setName(intValue);
        } else {
            LOG.log(Level.WARNING, "attempt to setNameFromAttribute with non existing attributeName {0}", attributeName);
        }
    }

    /**
     * Set the value of the TreatmentZone based on a attribute that should be
     * present in the IntegerAttributeMap if the map contains attribute name
     * will set the value of this TreatMentZone to the value in the map if the
     * parsed attributeName is is not found in the map will do nothing and log a
     * warning
     *
     * @param attributeName the key in the attributeMap
     */
    public void setValueFromAttribute(String attributeName) {
        Double theValue = getIntegerAttributeMap().get(attributeName);
        setValue(theValue);
    }

    @Override
    public Geometry getBoundingBox() {
        return (this.geometry != null) ? geometry.getEnvelope() : null;
    }

    @Override
    public Image getIcon() {
        if (!isValid()) {
            return ImageUtilities.mergeImages(ICON, ImageUtilities.loadImage("nl/cloudfarming/client/field/model/exclamation.png"), 0, 0);
        } else if (isAutocorrected()) {
            return ImageUtilities.mergeImages(ICON, ImageUtilities.loadImage("nl/cloudfarming/client/field/model/exclamation2.png"), 0, 0);
        }
        return ICON;
    }

    @Override
    public String getIconLabel() {
        return getTooltipText();
    }

    public boolean isValid() {
        return autocorrected || toplogyValidationError == null;
    }

    /**
     * Returns whether auto-correction been applied to this shapes geometry.
     * 
     * If true, this shape is considered valid and the validationMessage 
     * will contain the errors of the original geometry.
     * 
     * @return 
     */
    public boolean isAutocorrected() {
        return autocorrected;
    }

    public String getValidationMessage() {
        return toplogyValidationError == null ? null : toplogyValidationError.getMessage();
    }
    
    /**
     * Type of a shape object Gives an indication about what kind of shape it
     * is, what the purpose is and where it needs to be stored eventually.
     *
     * @author Timon Veenstra
     */
    @Messages({
        "ImportType.skip=Skip"
    })
    @Deprecated
    public enum ImportType {

        /**
         * Skip this shape
         */
        SKIP("Skip"),
        /**
         * Field shape, Geometry will be a MultiPolygon
         */
        FIELD("Field"),
        /**
         * Treatment zone shape, Geometry will be a MultiPolygon
         */
        TREATMENT_ZONE("Treatment zone");//,
        /**
         * help line shape, Geometry will be a line
         */
        // TODO: Reimplement Helplines in appication, and restore the Helpline beneath.
        //HELPLINE("Helpline");
        private String displayName;

        ImportType(String displayName) {
            this.displayName = displayName;
        }

        public String getDisplayName() {
            return displayName;
        }

        @Override
        public String toString() {
            return displayName;
        }
    }
}
